from ...utils import indent_string
from . import mixins
from .base import BaseFunctionalTest


class BaseStdioTest(BaseFunctionalTest):
    def _check_program_stdout(self, stdout):
        return True

    def _analyze_program_output(self, stdout, stderr, returncode):
        return self._pass_or_fail(
            pass_condition=(self.normalize_output(self.expected_output) == self.normalize_output(stdout)),
            fail_message="\n\tExpected:\n%s\tGot:\n%s" % (indent_string(self.expected_output), indent_string(stdout))
        )

    def _create_process(self):
        process = super(BaseStdioTest, self)._create_process()
        process.stdin.write(self.input)

        return process


class StdioTextWhitespaceSensitiveTest(BaseStdioTest): pass
class StdioTextTest(mixins.NormalizeWhitespacesMixin, BaseStdioTest): pass
class StdioNumbersTest(mixins.NormalizeToLineEndingNumbersMixin, BaseStdioTest): pass



import tempfile


PROGRAM_NAME = 'program.exe' # Normalize the compiled program name, so it works on all platforms
PROGRAM_TEST_DIR = tempfile.mkdtemp(prefix='smoke_test')

COMPILER_NAME = 'gcc'
COMPILER_SETTINGS = ['-g', '-o', PROGRAM_NAME]
LINKER_SETTINGS = ['-lm']

PROGRAM_POLL_TIME = 0.1 # Poll the running program status every X seconds
PROGRAM_MAX_WAIT_TIME = 3 # If not yet finished, terminate the program after X seconds run-time

FIXTURES_DIR = 'fixtures'



import hashlib
import logging

from pycparser import c_ast, parse_file
from pycparserext.ext_c_parser import GnuCParser

from .base import BaseTest


class BaseSourceCodeAnalysisTest(BaseTest):
    _ast_cache = {}

    @property
    def ast(self):
        with open(self.source_code_filename) as fp:
            file_hash = hashlib.sha1(fp.read()).hexdigest()

        if file_hash not in self._ast_cache:
            self._ast_cache[file_hash] = parse_file(
                self.source_code_filename,
                use_cpp=True,
                parser=GnuCParser(),
            )

        return self._ast_cache[file_hash]

    def _setup(self, *args, **kwargs):
        super(BaseSourceCodeAnalysisTest, self)._setup(*args, **kwargs)
        self.ast # HACK: A NO-OP used to activate the `_ast_cache` while we're still in the `_setup` stage


class CountFunctionCallsTest(BaseSourceCodeAnalysisTest):
    class FuncCallVisitor(c_ast.NodeVisitor):
        def __init__(self, function_name):
            self.function_name = function_name
            self.count = 0

        def visit_FuncCall(self, node):
            if node.name.name == self.function_name:
                logging.debug("'%s' function call found at %s", self.function_name, node.name.coord)
                self.count += 1

            self.generic_visit(node)

    def _test(self):
        function_name = self.fixture_id
        expected = self.fixture_data

        visitor = self.FuncCallVisitor(self.fixture_id)
        visitor.visit(self.ast)
        got = visitor.count

        return self._pass_or_fail(
            pass_condition=(expected == got),
            fail_message="There should be %d '%s' function call(s) in the source code (%d found)" % (expected, function_name, got)
        )



import logging

from . import constants as c
from .fixtures import load_fixtures_from
from .tests import BaseTest
from .utils import cleanup_test_dir, compile_source_code


def smoke_test(filename, skip_test_dir_cleanup=False):
    results = {}
    if compile_source_code(filename):
        for fixture_name, fixture_id, fixture_data in load_fixtures_from(c.FIXTURES_DIR):
            test = BaseTest.coerce(fixture_name)
            results[test.test_id(fixture_id)] = test(filename, fixture_id, fixture_data)

    if not skip_test_dir_cleanup:
        cleanup_test_dir()

    return results



import re


class NormalizeWhitespacesMixin(object):
    def normalize_output(self, s):
        return ' '.join(s.split())


class NormalizeToLineEndingNumbersMixin(object):
    FLOAT_RE = re.compile(r'([-+]?[0-9]+\.?[0-9]*)[ \t]*$')

    def normalize_output(self, s):
        normalized_results = []
        for line in s.split('\n'):
            if line.split(): # Skip empty lines
                line = line.replace('\r', '') # Remove '\r' for Windows/*NIX portability
                match = self.FLOAT_RE.search(line)
                if match:
                    normalized_results.append(float(match.group(1)))

        return normalized_results



import fnmatch
import os

import yaml

from .tests import BaseTest


def find_filenames(path, filename_pattern):
    file_list = sorted(os.listdir(path))
    for name in fnmatch.filter(file_list, filename_pattern):
        yield os.path.join(path, name)

def get_fixture_name(filename):
    return os.path.splitext(os.path.basename(filename))[0]

def validate_fixture_filenames(filenames):
    for filename in filenames:
        BaseTest.coerce(get_fixture_name(filename))
        yield filename

def open_fixture_files(filenames):
    for filename in filenames:
        with open(filename, 'r') as fp:
            yield filename, yaml.load(fp)

def parse_fixture_files(fixture_files):
    for filename, fixtures in fixture_files:
        fixture_name = get_fixture_name(filename)
        for fixture_id, fixture_data in sorted(fixtures.items()):
            yield fixture_name, fixture_id, fixture_data

def load_fixtures_from(fixtures_dir):
    fixture_filenames = find_filenames(fixtures_dir, '*.yaml')
    fixture_filenames = validate_fixture_filenames(fixture_filenames)
    fixture_files = open_fixture_files(fixture_filenames)
    fixtures = parse_fixture_files(fixture_files)

    return fixtures



import argparse
import logging
import os

from . import __version__
from .main import smoke_test
from .utils import get_test_stats


def main():
    # Setup command line option parser
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'filename',
        help='Source code file to examine'
    )
    parser.add_argument(
        '-q',
        '--quiet',
        action='store_const',
        const=logging.WARN,
        dest='verbosity',
        help='Be quiet, show only warnings and errors'
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_const',
        const=logging.DEBUG,
        dest='verbosity',
        help='Be very verbose, show debug information'
    )
    parser.add_argument(
        '--version',
        action='version',
        version="%(prog)s " + __version__
    )
    args = parser.parse_args()

    # Configure logging
    log_level = args.verbosity or logging.INFO
    logging.basicConfig(level=log_level, format="%(asctime)s [%(levelname)s] %(message)s")

    results = smoke_test(args.filename)
    num_ok, num_failed, success_rate = get_test_stats(results)
    if log_level <= logging.INFO:
        logging.info("Test stats: ok = %d, failed = %d, success rate = %d %%", num_ok, num_failed, success_rate)
    else:
        print success_rate

if __name__ == '__main__':
    main()



from contextlib import contextmanager
import logging
import re

from friendly_name_mixin import FriendlyNameFromClassMixin
from simple_plugins import PluginMount


class BaseTest(FriendlyNameFromClassMixin):
    _fixture_name_re = re.compile('( Test$)')

    __metaclass__ = PluginMount

    class Meta:
        id_field = 'fixture_name'
        id_field_coerce = str

    def __str__(self):
        return self.fixture_name

    @property
    def fixture_name(self):
        if not hasattr(self, '_fixture_name'):
            base_name = self._fixture_name_re.sub('', self.name)
            self._fixture_name = '-'.join(base_name.lower().split())

        return self._fixture_name

    def test_id(self, fixture_id):
        return "%s:%s" % (self.fixture_name, fixture_id)

    def _format_message(self, base, extra=None):
        return base if not extra else "%s: %s" % (base, extra)

    def _pass(self, message=None):
        logging.info(self._format_message("PASSED test '%s'" % self.test_id(self.fixture_id), message))
        return True

    def _fail(self, message=None):
        logging.error(self._format_message("FAILED test '%s'" % self.test_id(self.fixture_id), message))
        return False

    def _pass_or_fail(self, pass_condition, pass_message=None, fail_message=None):
        return self._pass(pass_message) if pass_condition else self._fail(fail_message)

    def _setup(self, source_code_filename, fixture_id, fixture_data):
        logging.debug("Preparing the environment to run test '%s'...", self.test_id(fixture_id))

        self.source_code_filename = source_code_filename
        self.fixture_id = fixture_id
        self.fixture_data = fixture_data

    def _teardown(self):
        logging.debug("Cleaning up the environment after running test '%s'...", self.test_id(self.fixture_id))

    @contextmanager
    def _test_environment(self, source_code_filename, fixture_id, fixture_data):
        try:
            self._setup(source_code_filename, fixture_id, fixture_data)
            yield
        finally:
            self._teardown()

    def _test(self):
        raise NotImplementedError

    def test(self, source_code_filename, fixture_id, fixture_data):
        with self._test_environment(source_code_filename, fixture_id, fixture_data):
            return bool(self._test())

    def __call__(self, source_code_filename, fixture_id, fixture_data):
        return self.test(source_code_filename, fixture_id, fixture_data)



from datetime import datetime, timedelta
import logging
import os
from string import Template
import subprocess
from time import sleep

from ... import constants as c
from ...utils import indent_string
from ..base import BaseTest


class BaseFunctionalTest(BaseTest):
    def _setup(self, *args, **kwargs):
        super(BaseFunctionalTest, self)._setup(*args, **kwargs)

        self.args = self.fixture_data.get('args', '')
        self.args_template_context = {}

        self.input = self.fixture_data.get('input', '')

        self.expected_output = self.fixture_data.get('expected-output', '')
        self.expected_returncode = self.fixture_data.get('expected-returncode', 0)

    def normalize_output(self, s):
        return s

    def _check_program_stdout(self, stdout):
        raise NotImplementedError

    def _check_program_stderr(self, stderr):
        return stderr == ''

    def _check_program_returncode(self, returncode):
        return returncode == self.expected_returncode

    def _sanity_check(self, stdout, stderr, returncode):
        if self._check_program_returncode(returncode) and self._check_program_stderr(stderr) and self._check_program_stdout(stdout):
            return True

        return self._fail(
            "Program terminated with errors:\n\treturn code: %d (expected %d)\n\tstdout:\n%s\n\tstderr:\n%s" % (
                returncode, self.expected_returncode, indent_string(stdout), indent_string(stderr)
            )
        )

    def _analyze_program_output(self, stdout, stderr, returncode):
        raise NotImplementedError

    def _resolve_args(self):
        return Template(self.args).substitute(self.args_template_context)

    def _create_process(self):
        process = subprocess.Popen(
            [os.path.join(c.PROGRAM_TEST_DIR, c.PROGRAM_NAME)] + self._resolve_args().split(),
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            cwd=c.PROGRAM_TEST_DIR,
        )

        return process

    def _test(self):
        process = self._create_process()

        start_time = datetime.now()
        while process.poll() is None:
            if datetime.now() - start_time > timedelta(seconds=c.PROGRAM_MAX_WAIT_TIME):
                process.terminate()
                return self._fail("Program took more than %d seconds to complete!" % c.PROGRAM_MAX_WAIT_TIME)

            logging.debug("Waiting for the program to complete...")
            sleep(c.PROGRAM_POLL_TIME)

        stdout = process.stdout.read()
        stderr = process.stderr.read()
        returncode = process.returncode
        if self._sanity_check(stdout, stderr, returncode):
            return self._analyze_program_output(stdout, stderr, returncode)



from setuptools import find_packages, setup


setup(
    name='smoke_test',
    version='1.0.1',
    url='https://github.com/petarmaric/smoke_test',
    license='BSD',
    author='Petar Maric',
    author_email='petarmaric@uns.ac.rs',
    description='Console app and Python API for automated smoke testing',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Education',
        'Topic :: Utilities',
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'smoke_test=smoke_test.shell:main',
        ],
        'distutils.commands': [
            'bdist_pex = pex.commands.bdist_pex:bdist_pex',
        ],
    },
    install_requires=[
        'friendly_name_mixin~=1.0',
        'pycparser==2.17',
        'pycparserext==2016.2',
        'PyYAML~=3.11',
        'simple_plugins~=1.0',
    ],
    extras_require={
        'dev': [
            'pex~=1.6.0',
        ],
    },
)



import logging
import os

from ... import constants as c
from ...utils import indent_string
from . import mixins
from .base import BaseFunctionalTest


class BaseFileTest(BaseFunctionalTest):
    def _setup(self, *args, **kwargs):
        super(BaseFileTest, self)._setup(*args, **kwargs)

        self.input_filename = self.fixture_id + '.in'
        self.output_filename = self.fixture_id + '.out'

        self.args_template_context = {
            'input_filename': self.input_filename,
            'output_filename': self.output_filename,
        }

        self.original_filenames = os.listdir(c.PROGRAM_TEST_DIR)
        logging.debug("Recording the existing filenames before running test: %s", self.original_filenames)

        with open(os.path.join(c.PROGRAM_TEST_DIR, self.input_filename), 'w') as out:
            logging.debug("Copying input file contents to '%s'", self.input_filename)
            out.write(self.input)

    def _teardown(self):
        super(BaseFileTest, self)._teardown()

        for filename in set(os.listdir(c.PROGRAM_TEST_DIR)) - set(self.original_filenames):
            logging.debug("Cleaning up after running test, removing '%s'", filename)
            os.remove(os.path.join(c.PROGRAM_TEST_DIR, filename))


class BaseFileTextTest(BaseFileTest):
    def _check_program_stdout(self, stdout):
        return stdout == ''

    def _analyze_program_output(self, stdout, stderr, returncode):
        output_file_path = os.path.join(c.PROGRAM_TEST_DIR, self.output_filename)
        if not os.path.exists(output_file_path):
            return self._fail("Program failed to create '%s'" % self.output_filename)

        with open(output_file_path) as fp_got:
            got = fp_got.read()
            return self._pass_or_fail(
                pass_condition=(self.normalize_output(self.expected_output) == self.normalize_output(got)),
                fail_message="\n\tExpected:\n%s\tGot:\n%s" % (indent_string(self.expected_output), indent_string(got))
            )


class FileTextWhitespaceSensitiveTest(BaseFileTextTest): pass
class FileTextTest(mixins.NormalizeWhitespacesMixin, BaseFileTextTest): pass
class FileNumbersTest(mixins.NormalizeToLineEndingNumbersMixin, BaseFileTextTest): pass


class BaseFileErrorHandlingTest(BaseFileTest):
    def _check_program_stdout(self, stdout):
        return True

    def _analyze_program_output(self, stdout, stderr, returncode):
        return self._pass()


class FileErrorInputNotReadableTest(BaseFileErrorHandlingTest): pass
class FileErrorOutputNotWritableTest(BaseFileErrorHandlingTest): pass



import logging
import os
import re
import subprocess
import shutil

from . import constants as c


def indent_string(s, indent=12):
    """
    Add the given number of space characters to the beginning every
    non-blank line in `s`, and return the result.
    """
    # This regexp matches the start of non-blank lines:
    return re.sub('(?m)^(?!$)', indent*' ', s)

def compile_source_code(filename):
    shutil.copy(filename, c.PROGRAM_TEST_DIR)

    process = subprocess.Popen(
        [c.COMPILER_NAME] + c.COMPILER_SETTINGS + [os.path.basename(filename)] + c.LINKER_SETTINGS,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        cwd=c.PROGRAM_TEST_DIR,
    )
    process.wait()
    stdout = process.stdout.read()
    stderr = process.stderr.read()

    if process.returncode != 0 or stdout or stderr:
        logging.fatal(
            "Can't compile source code:\n\treturn code: %d\n\tstdout:\n%s\n\tstderr:\n%s",
            process.returncode, indent_string(stdout), indent_string(stderr)
        )
        return False
    else:
        logging.info('Program compiled without errors nor warnings')
        return True

def cleanup_test_dir():
    logging.debug("Deleting the temporary testing directory '%s'", c.PROGRAM_TEST_DIR)
    shutil.rmtree(c.PROGRAM_TEST_DIR)

def get_test_stats(results):
    if not results:
        return 0, 0, 0 # num_ok, num_failed, success_rate

    num_total = len(results)

    num_ok = sum(results.values())
    num_failed = num_total - num_ok
    success_rate = int(round(100.0 * num_ok / num_total))

    return num_ok, num_failed, success_rate
